@@ -1,12 +1,13 @@
 from __future__ import annotations
 
-import asyncio
 import logging
+import threading
+import warnings
 from abc import ABC, abstractmethod
 from collections.abc import Callable, Generator, Iterable
-from concurrent.futures import ThreadPoolExecutor
+from dataclasses import dataclass, field
 from enum import StrEnum, auto
-from typing import TYPE_CHECKING, Any, TypeVar
+from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar
 
 from griptape_nodes.exe_types.core_types import (
     BaseNodeElement,
@@ -30,19 +31,18 @@ from griptape_nodes.retained_mode.events.base_events import (
     ProgressEvent,
     RequestPayload,
 )
-from griptape_nodes.retained_mode.events.execution_events import (
-    NodeUnresolvedEvent,
-    ParameterValueUpdateEvent,
-)
 from griptape_nodes.retained_mode.events.parameter_events import (
     AddParameterToNodeRequest,
     RemoveElementEvent,
     RemoveParameterFromNodeRequest,
 )
 from griptape_nodes.traits.options import Options
+from griptape_nodes.utils import async_utils
 
 if TYPE_CHECKING:
+    from griptape_nodes.exe_types.connections import Connections
     from griptape_nodes.exe_types.core_types import NodeMessagePayload
+    from griptape_nodes.node_library.library_registry import LibraryNameAndVersion
 
 logger = logging.getLogger("griptape_nodes")
 
@@ -50,6 +50,55 @@ T = TypeVar("T")
 
 AsyncResult = Generator[Callable[[], T], T]
 
+LOCAL_EXECUTION = "Local Execution"
+PRIVATE_EXECUTION = "Private Execution"
+CONTROL_INPUT_PARAMETER = "Control Input Selection"
+
+
+class ImportDependency(NamedTuple):
+    """Import dependency specification for a node.
+
+    Attributes:
+        module: The module name to import
+        class_name: Optional class name to import from the module. If None, imports the entire module.
+    """
+
+    module: str
+    class_name: str | None = None
+
+
+@dataclass
+class NodeDependencies:
+    """Dependencies that a node has on external resources.
+
+    This class provides a way for nodes to declare their dependencies on workflows,
+    static files, Python imports, and libraries. This information can be used by the system
+    for workflow packaging, dependency resolution, and deployment planning.
+
+    Attributes:
+        referenced_workflows: Set of workflow names that this node references
+        static_files: Set of static file names that this node depends on
+        imports: Set of Python imports that this node requires
+        libraries: Set of library names and versions that this node uses
+    """
+
+    referenced_workflows: set[str] = field(default_factory=set)
+    static_files: set[str] = field(default_factory=set)
+    imports: set[ImportDependency] = field(default_factory=set)
+    libraries: set[LibraryNameAndVersion] = field(default_factory=set)
+
+    def aggregate_from(self, other: NodeDependencies) -> None:
+        """Aggregate dependencies from another NodeDependencies object into this one.
+
+        Args:
+            other: The NodeDependencies object to aggregate from
+        """
+        # Aggregate all dependency types - no None checks needed since we use default_factory=set
+        self.referenced_workflows.update(other.referenced_workflows)
+        self.static_files.update(other.static_files)
+        self.imports.update(other.imports)
+        self.libraries.update(other.libraries)
+
 
 class NodeResolutionState(StrEnum):
     """Possible states for a node during resolution."""
@@ -59,23 +108,41 @@ class NodeResolutionState(StrEnum):
     RESOLVED = auto()
 
 
+def get_library_names_with_publish_handlers() -> list[str]:
+    """Get names of all registered libraries that have PublishWorkflowRequest handlers."""
+    from griptape_nodes.retained_mode.events.workflow_events import PublishWorkflowRequest
+    from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
+
+    library_manager = GriptapeNodes.LibraryManager()
+    event_handlers = library_manager.get_registered_event_handlers(PublishWorkflowRequest)
+
+    # Always include "local" and "private" as the first options
+    library_names = [LOCAL_EXECUTION, PRIVATE_EXECUTION]
+
+    # Add all registered library names that can handle PublishWorkflowRequest
+    library_names.extend(sorted(event_handlers.keys()))
+
+    return library_names
+
+
 class BaseNode(ABC):
     # Owned by a flow
     name: str
     metadata: dict[Any, Any]
 
     # Node Context Fields
-    state: NodeResolutionState
     current_spotlight_parameter: Parameter | None = None
     parameter_values: dict[str, Any]
     parameter_output_values: TrackedParameterOutputValues
     stop_flow: bool = False
     root_ui_element: BaseNodeElement
+    _state: NodeResolutionState
     _tracked_parameters: list[BaseNodeElement]
     _entry_control_parameter: Parameter | None = (
         None  # The control input parameter used to enter this node during execution
     )
     lock: bool = False  # When lock is true, the node is locked and can't be modified. When lock is false, the node is unlocked and can be modified.
+    _cancellation_requested: threading.Event  # Event indicating if cancellation has been requested for this node
 
     @property
     def parameters(self) -> list[Parameter]:
@@ -91,7 +158,7 @@ class BaseNode(ABC):
         state: NodeResolutionState = NodeResolutionState.UNRESOLVED,
     ) -> None:
         self.name = name
-        self.state = state
+        self._state = state
         if metadata is None:
             self.metadata = {}
         else:
@@ -103,7 +170,39 @@ class BaseNode(ABC):
         self.root_ui_element._node_context = self
         self.process_generator = None
         self._tracked_parameters = []
+        self._cancellation_requested = threading.Event()
         self.set_entry_control_parameter(None)
+        self.execution_environment = Parameter(
+            name="execution_environment",
+            tooltip="Environment that the node should execute in",
+            type=ParameterTypeBuiltin.STR,
+            allowed_modes={ParameterMode.PROPERTY},
+            default_value=LOCAL_EXECUTION,
+            traits={Options(choices=get_library_names_with_publish_handlers())},
+            ui_options={"hide": True},
+        )
+        self.add_parameter(self.execution_environment)
+        self.node_group = Parameter(
+            name="job_group",
+            tooltip="Groupings of multiple nodes to send up as a Deadline Cloud job.",
+            type=ParameterTypeBuiltin.STR,
+            allowed_modes={ParameterMode.PROPERTY},
+            default_value="",
+            ui_options={"hide": True},
+        )
+        self.add_parameter(self.node_group)
+
+    @property
+    def state(self) -> NodeResolutionState:
+        """Get the current resolution state of the node.
+
+        Existence as @property facilitates subclasses overriding the getter for dynamic/computed state.
+        """
+        return self._state
+
+    @state.setter
+    def state(self, new_state: NodeResolutionState) -> None:
+        self._state = new_state
 
     # This is gross and we need to have a universal pass on resolution state changes and emission of events. That's what this ticket does!
     # https://github.com/griptape-ai/griptape-nodes/issues/994
@@ -114,6 +213,7 @@ class BaseNode(ABC):
         if current_states_to_trigger_change_event is not None and self.state in current_states_to_trigger_change_event:
             # Trigger the change event.
             # Send an event to the GUI so it knows this node has changed resolution state.
+            from griptape_nodes.retained_mode.events.execution_events import NodeUnresolvedEvent
 
             GriptapeNodes.EventManager().put_event(
                 ExecutionGriptapeNodeEvent(
@@ -134,6 +234,27 @@ class BaseNode(ABC):
         """
         self._entry_control_parameter = parameter
 
+    @property
+    def is_cancellation_requested(self) -> bool:
+        """Check if cancellation has been requested for this node.
+
+        Returns:
+            True if cancellation has been requested, False otherwise
+        """
+        return self._cancellation_requested.is_set()
+
+    def request_cancellation(self) -> None:
+        """Request cancellation of this node's execution.
+
+        Sets a flag that the node can check during long-running operations
+        to cooperatively cancel execution.
+        """
+        self._cancellation_requested.set()
+
+    def clear_cancellation(self) -> None:
+        """Clear the cancellation request flag."""
+        self._cancellation_requested.clear()
+
     def emit_parameter_changes(self) -> None:
         if self._tracked_parameters:
             for parameter in self._tracked_parameters:
@@ -337,7 +458,13 @@ class BaseNode(ABC):
         if self.does_name_exist(param.name):
             msg = f"Cannot have duplicate names on parameters. Encountered two instances of '{param.name}'."
             raise ValueError(msg)
-        self.add_node_element(param)
+        parameter_group = (
+            self.get_group_by_name_or_element_id(param.parent_element_name) if param.parent_element_name else None
+        )
+        if parameter_group is not None:
+            parameter_group.add_child(param)
+        else:
+            self.add_node_element(param)
         self._emit_parameter_lifecycle_event(param)
 
     def remove_parameter_element_by_name(self, element_name: str) -> None:
@@ -384,9 +511,7 @@ class BaseNode(ABC):
         for name in names:
             parameter = self.get_parameter_by_name(name)
             if parameter is not None:
-                ui_options = parameter.ui_options
-                ui_options["hide"] = not visible
-                parameter.ui_options = ui_options
+                parameter.ui_options = {**parameter.ui_options, "hide": not visible}
 
     def get_message_by_name_or_element_id(self, element: str) -> ParameterMessage | None:
         element_items = self.root_ui_element.find_elements_by_type(ParameterMessage)
@@ -408,9 +533,7 @@ class BaseNode(ABC):
         for name in names:
             message = self.get_message_by_name_or_element_id(name)
             if message is not None:
-                ui_options = message.ui_options
-                ui_options["hide"] = not visible
-                message.ui_options = ui_options
+                message.ui_options = {**message.ui_options, "hide": not visible}
 
     def hide_message_by_name(self, names: str | list[str]) -> None:
         self._set_message_visibility(names, visible=False)
@@ -670,13 +793,15 @@ class BaseNode(ABC):
 
     def get_parameter_value(self, param_name: str) -> Any:
         param = self.get_parameter_by_name(param_name)
-        if param and isinstance(param, ParameterContainer):
+        if param is None:
+            return None
+        if isinstance(param, ParameterContainer):
             value = handle_container_parameter(self, param)
             if value is not None:
                 return value
         if param_name in self.parameter_values:
             return self.parameter_values[param_name]
-        return param.default_value if param else None
+        return param.default_value
 
     def get_parameter_list_value(self, param: str) -> list:
         """Flattens the given param from self.params into a single list.
@@ -727,11 +852,9 @@ class BaseNode(ABC):
                 return param
         return None
 
-    # Abstract method to process the node. Must be defined by the type
     # Must save the values of the output parameters in NodeContext.
-    @abstractmethod
-    def process[T](self) -> AsyncResult | None:
-        pass
+    def process(self) -> AsyncResult | None:
+        raise NotImplementedError
 
     async def aprocess(self) -> None:
         """Async version of process().
@@ -746,20 +869,14 @@ class BaseNode(ABC):
             return
 
         if isinstance(result, Generator):
-            # Handle generator pattern asynchronously using the same logic as before
-            loop = asyncio.get_running_loop()
-
             try:
                 # Start the generator
                 func = next(result)
 
                 while True:
-                    # Run callable in thread pool (preserving existing behavior)
-                    with ThreadPoolExecutor() as executor:
-                        future_result = await loop.run_in_executor(executor, func)
-
                     # Send result back and get next callable
-                    func = result.send(future_result)
+                    func_result = await async_utils.to_thread(func)
+                    func = result.send(func_result)
 
             except StopIteration:
                 # Generator is done
@@ -801,12 +918,26 @@ class BaseNode(ABC):
     def get_config_value(self, service: str, value: str) -> str:
         from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
 
+        warnings.warn(
+            "get_config_value() is deprecated. Use GriptapeNodes.SecretsManager().get_secret() for secrets/API keys "
+            "or GriptapeNodes.ConfigManager().get_config_value() for other config values.",
+            UserWarning,
+            stacklevel=2,
+        )
+
         config_value = GriptapeNodes.ConfigManager().get_config_value(f"nodes.{service}.{value}")
         return config_value
 
     def set_config_value(self, service: str, value: str, new_value: str) -> None:
         from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
 
+        warnings.warn(
+            "set_config_value() is deprecated. Use GriptapeNodes.SecretsManager().set_secret() for secrets/API keys "
+            "or GriptapeNodes.ConfigManager().set_config_value() for other config values.",
+            UserWarning,
+            stacklevel=2,
+        )
+
         GriptapeNodes.ConfigManager().set_config_value(f"nodes.{service}.{value}", new_value)
 
     def clear_node(self) -> None:
@@ -814,6 +945,8 @@ class BaseNode(ABC):
         self.state = NodeResolutionState.UNRESOLVED
         # delete all output values potentially generated
         self.parameter_output_values.clear()
+        # Clear cancellation flag
+        self.clear_cancellation()
         # Clear the spotlight linked list
         # First, clear all next/prev pointers to break the linked list
         current = self.current_spotlight_parameter
@@ -825,6 +958,35 @@ class BaseNode(ABC):
         # Then clear the reference to the first spotlight parameter
         self.current_spotlight_parameter = None
 
+    def get_node_dependencies(self) -> NodeDependencies | None:
+        """Return the dependencies that this node has on external resources.
+
+        This method should be overridden by nodes that have dependencies on:
+        - Referenced workflows: Other workflows that this node calls or references
+        - Static files: Files that this node reads from or requires for operation
+        - Python imports: Modules or classes that this node imports beyond standard dependencies
+
+        This information can be used by the system for workflow packaging, dependency
+        resolution, deployment planning, and ensuring all required resources are available.
+
+        Returns:
+            NodeDependencies object containing the node's dependencies, or None if the node
+            has no external dependencies beyond the standard framework dependencies.
+
+        Example:
+            def get_node_dependencies(self) -> NodeDependencies | None:
+                return NodeDependencies(
+                    referenced_workflows={"image_processing_workflow", "validation_workflow"},
+                    static_files={"config.json", "model_weights.pkl"},
+                    imports={
+                        ImportDependency("numpy"),
+                        ImportDependency("sklearn.linear_model", "LinearRegression"),
+                        ImportDependency("custom_module", "SpecialProcessor")
+                    }
+                )
+        """
+        return None
+
     def append_value_to_parameter(self, parameter_name: str, value: Any) -> None:
         from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
 
@@ -847,6 +1009,7 @@ class BaseNode(ABC):
         )
 
     def publish_update_to_parameter(self, parameter_name: str, value: Any) -> None:
+        from griptape_nodes.retained_mode.events.execution_events import ParameterValueUpdateEvent
         from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
 
         parameter = self.get_parameter_by_name(parameter_name)
@@ -867,7 +1030,7 @@ class BaseNode(ABC):
             msg = f"Parameter '{parameter_name} doesn't exist on {self.name}'"
             raise RuntimeError(msg)
 
-    def reorder_elements(self, element_order: list[str | int]) -> None:
+    def reorder_elements(self, element_order: list[str] | list[int] | list[str | int]) -> None:
         """Reorder the elements of this node.
 
         Args:
@@ -909,7 +1072,10 @@ class BaseNode(ABC):
 
         # Verify we have all elements
         if len(ordered_elements) != len(current_elements):
-            msg = "Element order must include all elements exactly once"
+            ordered_names = {e.name for e in ordered_elements}
+            current_names = {e.name for e in current_elements}
+            diff = current_names - ordered_names
+            msg = f"Element order must include all elements exactly once. Missing from new order: {diff}"
             raise ValueError(msg)
 
         # Remove all elements from root_ui_element
@@ -1083,6 +1249,47 @@ class BaseNode(ABC):
         # Use reorder_elements to apply the move
         self.reorder_elements(list(new_order))
 
+    def get_element_index(self, element: str | BaseNodeElement, root: BaseNodeElement | None = None) -> int:
+        """Get the current index of an element in the element list.
+
+        Args:
+            element: The element to get the index for, specified by name or element object
+            root: The root element to search within. If None, uses root_ui_element
+
+        Returns:
+            The current index of the element (0-based)
+
+        Raises:
+            ValueError: If element is not found
+
+        Example:
+            # Get index by name in root container
+            index = node.get_element_index("element1")
+
+            # Get index within a specific parameter group
+            group = node.get_element_by_name_and_type("my_group", ParameterGroup)
+            index = node.get_element_index("parameter1", root=group)
+
+            # Get index of a parameter to position another element relative to it
+            reference_index = node.get_element_index("some_parameter")
+            node.move_element_to_position("new_parameter", reference_index + 1)
+        """
+        # Use root_ui_element if no root specified
+        if root is None:
+            root = self.root_ui_element
+
+        # Get list of all element names in the root
+        element_names = [child.name for child in root._children]
+
+        # Get element name
+        if isinstance(element, str):
+            element_name = element
+        else:
+            element_name = element.name
+
+        # Find the index of the element
+        return element_names.index(element_name)
+
 
 class TrackedParameterOutputValues(dict[str, Any]):
     """A dictionary that tracks modifications and emits AlterElementEvent when parameter output values change."""
@@ -1109,7 +1316,10 @@ class TrackedParameterOutputValues(dict[str, Any]):
             keys_to_clear = list(self.keys())
             super().clear()
             for key in keys_to_clear:
-                self._emit_parameter_change_event(key, None, deleted=True)
+                # Some nodes still have values set, even if their output values are cleared
+                # Here, we are emitting an event with those set values, to not misrepresent the values of the parameters in the UI.
+                value = self._node.get_parameter_value(key)
+                self._emit_parameter_change_event(key, value, deleted=True)
 
     def silent_clear(self) -> None:
         """Clear all values without emitting parameter change events."""
@@ -1374,13 +1584,16 @@ class EndNode(BaseNode):
             case self.succeeded_control:
                 was_successful = True
                 status_prefix = "[SUCCEEDED]"
+                logger.debug("End Node '%s': Matched succeeded_control path", self.name)
             case self.failed_control:
                 was_successful = False
                 status_prefix = "[FAILED]"
+                logger.debug("End Node '%s': Matched failed_control path", self.name)
             case _:
                 # No specific success/failure connection provided, assume success
                 was_successful = True
                 status_prefix = "[SUCCEEDED] No connection provided for success or failure, assuming successful"
+                logger.debug("End Node '%s': No specific control connection, assuming success", self.name)
 
         # Get result details and format the final message
         result_details_value = self.get_parameter_value("result_details")
@@ -1393,6 +1606,16 @@ class EndNode(BaseNode):
 
         self.status_component.set_execution_result(was_successful=was_successful, result_details=details)
 
+        # Update all values to use the output value
+        for param in self.parameters:
+            if param.type != ParameterTypeBuiltin.CONTROL_TYPE:
+                value = self.get_parameter_value(param.name)
+                self.parameter_output_values[param.name] = value
+        entry_parameter = self._entry_control_parameter
+        # Update which control parameter to flag as the output value.
+        if entry_parameter is not None:
+            self.parameter_output_values[entry_parameter.name] = CONTROL_INPUT_PARAMETER
+
 
 class StartLoopNode(BaseNode):
     end_node: EndLoopNode | None = None
@@ -1624,6 +1847,268 @@ class ErrorProxyNode(BaseNode):
         return None
 
 
+@dataclass
+class NodeGroup:
+    """Represents a group of nodes that should be executed together in parallel.
+
+    Nodes in a group are identified by having the same non-empty value in their
+    node_group parameter. During DAG resolution, grouped nodes are replaced with
+    a single NodeGroupProxyNode that represents them in the execution graph.
+
+    Attributes:
+        group_id: Unique identifier for this group (value from node_group parameter)
+        nodes: Set of BaseNode instances that belong to this group
+        internal_connections: Connections between nodes within the group
+        external_incoming_connections: Connections from outside nodes into the group
+        external_outgoing_connections: Connections from group nodes to outside nodes
+    """
+
+    group_id: str
+    nodes: dict[str, BaseNode] = field(default_factory=dict)
+    internal_connections: list[Connection] = field(default_factory=list)
+    external_incoming_connections: list[Connection] = field(default_factory=list)
+    external_outgoing_connections: list[Connection] = field(default_factory=list)
+    # Store original node references before remapping to proxy (for cleanup)
+    original_incoming_targets: dict[int, BaseNode] = field(default_factory=dict)  # conn_id -> original target
+    original_outgoing_sources: dict[int, BaseNode] = field(default_factory=dict)  # conn_id -> original source
+
+    def add_node(self, node: BaseNode) -> None:
+        """Add a node to this group."""
+        self.nodes[node.name] = node
+
+    def validate_no_intermediate_nodes(self, all_connections: dict[int, Connection]) -> None:
+        """Validate that no ungrouped nodes exist between grouped nodes.
+
+        This method checks the dependency graph to ensure that all nodes that lie
+        on paths between grouped nodes are also part of the group. If ungrouped
+        nodes are found between grouped nodes, this indicates a logical error in
+        the group definition.
+
+        Args:
+            all_connections: Dictionary mapping connection IDs to Connection objects
+
+        Raises:
+            ValueError: If ungrouped nodes are found between grouped nodes
+        """
+        from griptape_nodes.exe_types.connections import Connections
+
+        # Build a Connections object for traversal
+        connections = Connections()
+        connections.connections = all_connections
+
+        # Rebuild indices for efficient lookup
+        for conn_id, conn in all_connections.items():
+            connections.outgoing_index.setdefault(conn.source_node.name, {}).setdefault(
+                conn.source_parameter.name, []
+            ).append(conn_id)
+            connections.incoming_index.setdefault(conn.target_node.name, {}).setdefault(
+                conn.target_parameter.name, []
+            ).append(conn_id)
+
+        # Check each pair of nodes in the group
+        for node_a in self.nodes.values():
+            for node_b in self.nodes.values():
+                if node_a == node_b:
+                    continue
+
+                # Check if there's a path from node_a to node_b
+                intermediate_nodes = self._find_intermediate_nodes(node_a, node_b, connections)
+
+                # Check if any intermediate nodes are not in the group
+                ungrouped_intermediates = [n for n in intermediate_nodes if n not in self.nodes]
+
+                if ungrouped_intermediates:
+                    ungrouped_names = [n.name for n in ungrouped_intermediates]
+                    msg = (
+                        f"Invalid node group '{self.group_id}': Found ungrouped nodes between grouped nodes. "
+                        f"Ungrouped nodes {ungrouped_names} exist on the path from '{node_a.name}' to '{node_b.name}'. "
+                        f"All nodes on paths between grouped nodes must be part of the same group."
+                    )
+                    raise ValueError(msg)
+
+    def _find_intermediate_nodes(  # noqa: C901
+        self, start_node: BaseNode, end_node: BaseNode, connections: Connections
+    ) -> set[BaseNode]:
+        """Find all nodes on paths between start_node and end_node (excluding endpoints).
+
+        Uses BFS to explore all paths from start_node to end_node and collects
+        all nodes encountered (except start and end nodes themselves).
+
+        Args:
+            start_node: Starting node for path search
+            end_node: Target node for path search
+            connections: Connections object for graph traversal
+
+        Returns:
+            Set of nodes found on paths between start and end (excluding endpoints)
+        """
+        if start_node.name not in connections.outgoing_index:
+            return set()
+
+        visited = set()
+        intermediate = set()
+        queue = [(start_node, [start_node])]
+
+        while queue:
+            current_node, path = queue.pop(0)
+
+            if current_node.name in visited:
+                continue
+            visited.add(current_node.name)
+
+            # Process outgoing connections from current node
+            if current_node.name not in connections.outgoing_index:
+                continue
+
+            for conn_ids in connections.outgoing_index[current_node.name].values():
+                for conn_id in conn_ids:
+                    if conn_id not in connections.connections:
+                        continue
+
+                    conn = connections.connections[conn_id]
+                    next_node = conn.target_node
+
+                    # If we reached the end node, record intermediate nodes
+                    if next_node == end_node:
+                        for node in path[1:]:
+                            intermediate.add(node)
+                        continue
+
+                    # Continue exploring if not already visited
+                    if next_node.name not in visited:
+                        queue.append((next_node, [*path, next_node]))
+
+        return intermediate
+
+
+class NodeGroupProxyNode(BaseNode):
+    """Proxy node that represents a group of nodes during DAG execution.
+
+    This node acts as a single execution unit for a group of nodes that should
+    be executed in parallel. When the DAG executor encounters this proxy node,
+    it passes the entire NodeGroup to the NodeExecutor which handles parallel
+    execution of all grouped nodes.
+
+    The proxy node has parameters that mirror the external connections to/from
+    the group, allowing it to seamlessly integrate into the DAG structure.
+
+    Attributes:
+        node_group: The NodeGroup instance this proxy represents
+    """
+
+    def __init__(
+        self,
+        name: str,
+        node_group: NodeGroup,
+        metadata: dict[Any, Any] | None = None,
+    ) -> None:
+        super().__init__(name, metadata)
+        self.node_group_data = node_group
+
+        # Track mapping from proxy parameter name to (original_node, original_param_name)
+        self._proxy_param_to_node_param: dict[str, tuple[BaseNode, str]] = {}
+        execution_type = set()
+        for node in node_group.nodes.values():
+            execution_type.add(node.get_parameter_value(node.execution_environment.name))
+        # TODO: Set this by group in the UI, not set on the node itself. https://github.com/griptape-ai/griptape-vsl-gui/issues/1429
+        if len(execution_type) > 1:
+            # Hoping this check can be removed by UI updates.
+            # For now, we are setting execution type individually on a parameters in all three of the nodes. we want their execution types to all be matching, or we fail.
+            msg = f"Node group '{node_group.group_id}' has nodes with multiple execution types: {execution_type}"
+            raise ValueError(msg)
+        self.set_parameter_value(self.execution_environment.name, execution_type.pop())
+        # Note: Proxy parameters are created AFTER connection remapping in control_flow.py
+        # via explicit call to create_proxy_parameters()
+
+    def create_proxy_parameters(self) -> None:
+        """Create parameters on the proxy that match external connections.
+
+        For each external incoming connection, create an input parameter with name
+        format: {sanitized_node_name}__{param_name}. This allows the proxy to
+        forward parameter values to the correct original node.
+
+        For each external outgoing connection, create an output parameter.
+        This allows the proxy to integrate seamlessly into the DAG.
+        """
+        # Track created parameters to avoid duplicates
+        created_params = set()
+
+        # Create input parameters for external incoming connections
+        for conn in self.node_group_data.external_incoming_connections:
+            conn_id = id(conn)
+            # Get the original target node from saved mapping (before it was remapped to proxy)
+            target_node = self.node_group_data.original_incoming_targets.get(conn_id)
+            if target_node is None:
+                # Fallback if not found (shouldn't happen)
+                msg = f"Failed to find target node for incoming connection with ID: {conn_id}"
+                raise ValueError(msg)
+            target_param = conn.target_parameter
+
+            # Create proxy parameter name: {sanitized_node_name}__{param_name}
+            sanitized_node_name = target_node.name.replace(" ", "_")
+            proxy_param_name = f"{sanitized_node_name}__{target_param.name}"
+
+            if proxy_param_name not in created_params:
+                proxy_param = Parameter(
+                    name=proxy_param_name,
+                    type=target_param.type,
+                    input_types=target_param.input_types,
+                    output_type=target_param.output_type,
+                    tooltip=f"Proxy input for {target_node.name}.{target_param.name}",
+                    allowed_modes={ParameterMode.INPUT},
+                )
+                self.add_parameter(proxy_param)
+                created_params.add(proxy_param_name)
+
+                # Track mapping from proxy param to original node/param
+                self._proxy_param_to_node_param[proxy_param_name] = (target_node, target_param.name)
+
+        # Create output parameters for external outgoing connections
+        for conn in self.node_group_data.external_outgoing_connections:
+            conn_id = id(conn)
+            # Get the original source node from saved mapping (before it was remapped to proxy)
+            source_node = self.node_group_data.original_outgoing_sources.get(conn_id)
+            if source_node is None:
+                # Fallback if not found (shouldn't happen)
+                continue
+
+            source_param = conn.source_parameter
+
+            # Create proxy parameter name: {sanitized_node_name}__{param_name}
+            sanitized_node_name = source_node.name.replace(" ", "_")
+            proxy_param_name = f"{sanitized_node_name}__{source_param.name}"
+
+            if proxy_param_name not in created_params:
+                proxy_param = Parameter(
+                    name=proxy_param_name,
+                    type=source_param.type,
+                    input_types=source_param.input_types,
+                    output_type=source_param.output_type,
+                    tooltip=f"Proxy output for {source_node.name}.{source_param.name}",
+                    allowed_modes={ParameterMode.OUTPUT},
+                )
+                self.add_parameter(proxy_param)
+                created_params.add(proxy_param_name)
+
+                # Track mapping from proxy param to original node/param
+                self._proxy_param_to_node_param[proxy_param_name] = (source_node, source_param.name)
+
+    async def aprocess(self) -> None:
+        """Execute all nodes in the group in parallel.
+
+        This method is called by the DAG executor. It executes all nodes in the
+        group concurrently using asyncio.gather and handles propagating input
+        values from the proxy to the grouped nodes.
+        """
+        msg = "NodeGroupProxyNode should not be executed locally."
+        raise NotImplementedError(msg)
+
+    def process(self) -> Any:
+        """Synchronous process method - not used for proxy nodes."""
+        msg = "NodeGroupProxyNode should use aprocess() for async execution."
+        raise NotImplementedError(msg)
+
+
 class Connection:
     source_node: BaseNode
     target_node: BaseNode
