@@ -2,42 +2,23 @@ from __future__ import annotations
 
 import logging
 import os
-import re
-import uuid
-from dataclasses import dataclass
 from datetime import UTC, datetime
 from typing import TYPE_CHECKING, Any
 
+import semver
+
 from griptape_nodes.exe_types.flow import ControlFlow
 from griptape_nodes.node_library.workflow_registry import WorkflowRegistry
 from griptape_nodes.retained_mode.events.app_events import (
-    AppConnectionEstablished,
-    AppEndSessionRequest,
-    AppEndSessionResultFailure,
-    AppEndSessionResultSuccess,
-    AppGetSessionRequest,
-    AppGetSessionResultSuccess,
-    AppStartSessionRequest,
-    AppStartSessionResultSuccess,
     EngineHeartbeatRequest,
     EngineHeartbeatResultFailure,
     EngineHeartbeatResultSuccess,
-    GetEngineNameRequest,
-    GetEngineNameResultFailure,
-    GetEngineNameResultSuccess,
     GetEngineVersionRequest,
     GetEngineVersionResultFailure,
     GetEngineVersionResultSuccess,
-    SessionHeartbeatRequest,
-    SessionHeartbeatResultFailure,
-    SessionHeartbeatResultSuccess,
-    SetEngineNameRequest,
-    SetEngineNameResultFailure,
-    SetEngineNameResultSuccess,
 )
 from griptape_nodes.retained_mode.events.base_events import (
     GriptapeNodeEvent,
-    ResultDetails,
     ResultPayloadFailure,
 )
 from griptape_nodes.retained_mode.events.flow_events import (
@@ -62,6 +43,7 @@ if TYPE_CHECKING:
     from griptape_nodes.retained_mode.managers.event_manager import EventManager
     from griptape_nodes.retained_mode.managers.flow_manager import FlowManager
     from griptape_nodes.retained_mode.managers.library_manager import LibraryManager
+    from griptape_nodes.retained_mode.managers.mcp_manager import MCPManager
     from griptape_nodes.retained_mode.managers.model_manager import ModelManager
     from griptape_nodes.retained_mode.managers.node_manager import NodeManager
     from griptape_nodes.retained_mode.managers.object_manager import ObjectManager
@@ -69,12 +51,15 @@ if TYPE_CHECKING:
         OperationDepthManager,
     )
     from griptape_nodes.retained_mode.managers.os_manager import OSManager
+    from griptape_nodes.retained_mode.managers.project_manager import ProjectManager
+    from griptape_nodes.retained_mode.managers.resource_manager import ResourceManager
     from griptape_nodes.retained_mode.managers.secrets_manager import SecretsManager
     from griptape_nodes.retained_mode.managers.session_manager import SessionManager
     from griptape_nodes.retained_mode.managers.static_files_manager import (
         StaticFilesManager,
     )
     from griptape_nodes.retained_mode.managers.sync_manager import SyncManager
+    from griptape_nodes.retained_mode.managers.user_manager import UserManager
     from griptape_nodes.retained_mode.managers.variable_manager import (
         VariablesManager,
     )
@@ -87,48 +72,6 @@ if TYPE_CHECKING:
 logger = logging.getLogger("griptape_nodes")
 
 
-@dataclass
-class Version:
-    major: int
-    minor: int
-    patch: int
-
-    @classmethod
-    def from_string(cls, version_string: str) -> Version | None:
-        match = re.match(r"(\d+)\.(\d+)\.(\d+)", version_string)
-        if match:
-            major, minor, patch = map(int, match.groups())
-            return cls(major, minor, patch)
-        return None
-
-    def __str__(self) -> str:
-        return f"{self.major}.{self.minor}.{self.patch}"
-
-    def __lt__(self, other: Version) -> bool:
-        """Less than comparison."""
-        return (self.major, self.minor, self.patch) < (other.major, other.minor, other.patch)
-
-    def __le__(self, other: Version) -> bool:
-        """Less than or equal comparison."""
-        return (self.major, self.minor, self.patch) <= (other.major, other.minor, other.patch)
-
-    def __gt__(self, other: Version) -> bool:
-        """Greater than comparison."""
-        return (self.major, self.minor, self.patch) > (other.major, other.minor, other.patch)
-
-    def __ge__(self, other: Version) -> bool:
-        """Greater than or equal comparison."""
-        return (self.major, self.minor, self.patch) >= (other.major, other.minor, other.patch)
-
-    def __eq__(self, other: Version) -> bool:  # type: ignore[override]
-        """Equality comparison."""
-        return (self.major, self.minor, self.patch) == (other.major, other.minor, other.patch)
-
-    def __hash__(self) -> int:
-        """Hash function for Version."""
-        return hash((self.major, self.minor, self.patch))
-
-
 class GriptapeNodes(metaclass=SingletonMeta):
     _event_manager: EventManager
     _os_manager: OSManager
@@ -149,9 +92,13 @@ class GriptapeNodes(metaclass=SingletonMeta):
     _version_compatibility_manager: VersionCompatibilityManager
     _session_manager: SessionManager
     _engine_identity_manager: EngineIdentityManager
+    _mcp_manager: MCPManager
+    _resource_manager: ResourceManager
     _sync_manager: SyncManager
+    _user_manager: UserManager
+    _project_manager: ProjectManager
 
-    def __init__(self) -> None:
+    def __init__(self) -> None:  # noqa: PLR0915
         from griptape_nodes.retained_mode.managers.agent_manager import AgentManager
         from griptape_nodes.retained_mode.managers.arbitrary_code_exec_manager import (
             ArbitraryCodeExecManager,
@@ -162,6 +109,7 @@ class GriptapeNodes(metaclass=SingletonMeta):
         from griptape_nodes.retained_mode.managers.event_manager import EventManager
         from griptape_nodes.retained_mode.managers.flow_manager import FlowManager
         from griptape_nodes.retained_mode.managers.library_manager import LibraryManager
+        from griptape_nodes.retained_mode.managers.mcp_manager import MCPManager
         from griptape_nodes.retained_mode.managers.model_manager import ModelManager
         from griptape_nodes.retained_mode.managers.node_manager import NodeManager
         from griptape_nodes.retained_mode.managers.object_manager import ObjectManager
@@ -169,12 +117,15 @@ class GriptapeNodes(metaclass=SingletonMeta):
             OperationDepthManager,
         )
         from griptape_nodes.retained_mode.managers.os_manager import OSManager
+        from griptape_nodes.retained_mode.managers.project_manager import ProjectManager
+        from griptape_nodes.retained_mode.managers.resource_manager import ResourceManager
         from griptape_nodes.retained_mode.managers.secrets_manager import SecretsManager
         from griptape_nodes.retained_mode.managers.session_manager import SessionManager
         from griptape_nodes.retained_mode.managers.static_files_manager import (
             StaticFilesManager,
         )
         from griptape_nodes.retained_mode.managers.sync_manager import SyncManager
+        from griptape_nodes.retained_mode.managers.user_manager import UserManager
         from griptape_nodes.retained_mode.managers.variable_manager import (
             VariablesManager,
         )
@@ -188,6 +139,7 @@ class GriptapeNodes(metaclass=SingletonMeta):
         # Initialize only if our managers haven't been created yet
         if not hasattr(self, "_event_manager"):
             self._event_manager = EventManager()
+            self._resource_manager = ResourceManager(self._event_manager)
             self._config_manager = ConfigManager(self._event_manager)
             self._os_manager = OSManager(self._event_manager)
             self._secrets_manager = SecretsManager(self._config_manager, self._event_manager)
@@ -206,32 +158,20 @@ class GriptapeNodes(metaclass=SingletonMeta):
             )
             self._agent_manager = AgentManager(self._static_files_manager, self._event_manager)
             self._version_compatibility_manager = VersionCompatibilityManager(self._event_manager)
-            self._session_manager = SessionManager(self._event_manager)
             self._engine_identity_manager = EngineIdentityManager(self._event_manager)
+            self._session_manager = SessionManager(self._engine_identity_manager, self._event_manager)
+            self._mcp_manager = MCPManager(self._event_manager, self._config_manager)
             self._sync_manager = SyncManager(self._event_manager, self._config_manager)
+            self._user_manager = UserManager(self._secrets_manager)
+            self._project_manager = ProjectManager(self._event_manager, self._config_manager, self._secrets_manager)
 
             # Assign handlers now that these are created.
             self._event_manager.assign_manager_to_request_type(
                 GetEngineVersionRequest, self.handle_engine_version_request
             )
-            self._event_manager.assign_manager_to_request_type(
-                AppStartSessionRequest, self.handle_session_start_request
-            )
-            self._event_manager.assign_manager_to_request_type(AppEndSessionRequest, self.handle_session_end_request)
-            self._event_manager.add_listener_to_app_event(AppConnectionEstablished, self.on_app_connection_established)
-            self._event_manager.assign_manager_to_request_type(AppGetSessionRequest, self.handle_get_session_request)
-            self._event_manager.assign_manager_to_request_type(
-                SessionHeartbeatRequest, self.handle_session_heartbeat_request
-            )
             self._event_manager.assign_manager_to_request_type(
                 EngineHeartbeatRequest, self.handle_engine_heartbeat_request
             )
-            self._event_manager.assign_manager_to_request_type(
-                GetEngineNameRequest, self.handle_get_engine_name_request
-            )
-            self._event_manager.assign_manager_to_request_type(
-                SetEngineNameRequest, self.handle_set_engine_name_request
-            )
 
     @classmethod
     def get_instance(cls) -> GriptapeNodes:
@@ -296,11 +236,11 @@ class GriptapeNodes(metaclass=SingletonMeta):
 
     @classmethod
     def get_session_id(cls) -> str | None:
-        return GriptapeNodes.SessionManager().get_active_session_id()
+        return GriptapeNodes.SessionManager().active_session_id
 
     @classmethod
     def get_engine_id(cls) -> str | None:
-        return GriptapeNodes.EngineIdentityManager().get_active_engine_id()
+        return GriptapeNodes.EngineIdentityManager().active_engine_id
 
     @classmethod
     def EventManager(cls) -> EventManager:
@@ -370,10 +310,18 @@ class GriptapeNodes(metaclass=SingletonMeta):
     def SessionManager(cls) -> SessionManager:
         return GriptapeNodes.get_instance()._session_manager
 
+    @classmethod
+    def MCPManager(cls) -> MCPManager:
+        return GriptapeNodes.get_instance()._mcp_manager
+
     @classmethod
     def EngineIdentityManager(cls) -> EngineIdentityManager:
         return GriptapeNodes.get_instance()._engine_identity_manager
 
+    @classmethod
+    def ResourceManager(cls) -> ResourceManager:
+        return GriptapeNodes.get_instance()._resource_manager
+
     @classmethod
     def SyncManager(cls) -> SyncManager:
         return GriptapeNodes.get_instance()._sync_manager
@@ -382,6 +330,14 @@ class GriptapeNodes(metaclass=SingletonMeta):
     def VariablesManager(cls) -> VariablesManager:
         return GriptapeNodes.get_instance()._workflow_variables_manager
 
+    @classmethod
+    def UserManager(cls) -> UserManager:
+        return GriptapeNodes.get_instance()._user_manager
+
+    @classmethod
+    def ProjectManager(cls) -> ProjectManager:
+        return GriptapeNodes.get_instance()._project_manager
+
     @classmethod
     def clear_data(cls) -> None:
         # Get canvas
@@ -405,113 +361,20 @@ class GriptapeNodes(metaclass=SingletonMeta):
             msg = "Failed to successfully delete all objects"
             raise ValueError(msg)
 
-    async def on_app_connection_established(self, _payload: AppConnectionEstablished) -> None:
-        from griptape_nodes.app.app import subscribe_to_topic
-
-        # Subscribe to request topic (engine discovery)
-        await subscribe_to_topic("request")
-
-        # Get engine ID and subscribe to engine_id/request
-        engine_id = GriptapeNodes.get_engine_id()
-        if engine_id:
-            await subscribe_to_topic(f"engines/{engine_id}/request")
-        else:
-            logger.warning("Engine ID not available for subscription")
-
-        # Get session ID and subscribe to session_id/request if available
-        session_id = GriptapeNodes.get_session_id()
-        if session_id:
-            topic = f"sessions/{session_id}/request"
-            await subscribe_to_topic(topic)
-        else:
-            logger.info("No session ID available for subscription")
-
     def handle_engine_version_request(self, request: GetEngineVersionRequest) -> ResultPayload:  # noqa: ARG002
         try:
-            engine_ver = Version.from_string(engine_version)
-            if engine_ver:
-                return GetEngineVersionResultSuccess(
-                    major=engine_ver.major,
-                    minor=engine_ver.minor,
-                    patch=engine_ver.patch,
-                    result_details="Engine version retrieved successfully.",
-                )
-            details = f"Attempted to get engine version. Failed because version string '{engine_ver}' wasn't in expected major.minor.patch format."
-            logger.error(details)
-            return GetEngineVersionResultFailure(result_details=details)
+            engine_ver = semver.VersionInfo.parse(engine_version)
+            return GetEngineVersionResultSuccess(
+                major=engine_ver.major,
+                minor=engine_ver.minor,
+                patch=engine_ver.patch,
+                result_details="Engine version retrieved successfully.",
+            )
         except Exception as err:
             details = f"Attempted to get engine version. Failed due to '{err}'."
             logger.error(details)
             return GetEngineVersionResultFailure(result_details=details)
 
-    async def handle_session_start_request(self, request: AppStartSessionRequest) -> ResultPayload:  # noqa: ARG002
-        from griptape_nodes.app.app import subscribe_to_topic
-
-        current_session_id = GriptapeNodes.SessionManager().get_active_session_id()
-        if current_session_id is None:
-            # Client wants a new session
-            current_session_id = uuid.uuid4().hex
-            GriptapeNodes.SessionManager().save_session(current_session_id)
-            details = f"New session '{current_session_id}' started at {datetime.now(tz=UTC)}."
-            logger.info(details)
-        else:
-            details = f"Session '{current_session_id}' already active. Joining..."
-
-        topic = f"sessions/{current_session_id}/request"
-        await subscribe_to_topic(topic)
-        logger.info("Subscribed to new session topic: %s", topic)
-
-        return AppStartSessionResultSuccess(current_session_id, result_details="Session started successfully.")
-
-    async def handle_session_end_request(self, _: AppEndSessionRequest) -> ResultPayload:
-        from griptape_nodes.app.app import unsubscribe_from_topic
-
-        try:
-            previous_session_id = GriptapeNodes.SessionManager().get_active_session_id()
-            if previous_session_id is None:
-                details = "No active session to end."
-                logger.info(details)
-            else:
-                details = f"Session '{previous_session_id}' ended at {datetime.now(tz=UTC)}."
-                logger.info(details)
-                GriptapeNodes.SessionManager().clear_saved_session()
-
-            unsubscribe_topic = f"sessions/{previous_session_id}/request"
-            await unsubscribe_from_topic(unsubscribe_topic)
-
-            return AppEndSessionResultSuccess(
-                session_id=previous_session_id, result_details="Session ended successfully."
-            )
-        except Exception as err:
-            details = f"Failed to end session due to '{err}'."
-            logger.error(details)
-            return AppEndSessionResultFailure(result_details=details)
-
-    def handle_get_session_request(self, _: AppGetSessionRequest) -> ResultPayload:
-        return AppGetSessionResultSuccess(
-            session_id=GriptapeNodes.SessionManager().get_active_session_id(),
-            result_details="Session ID retrieved successfully.",
-        )
-
-    def handle_session_heartbeat_request(self, request: SessionHeartbeatRequest) -> ResultPayload:  # noqa: ARG002
-        """Handle session heartbeat requests.
-
-        Simply verifies that the session is active and responds with success.
-        """
-        try:
-            active_session_id = GriptapeNodes.SessionManager().get_active_session_id()
-            if active_session_id is None:
-                details = "Session heartbeat received but no active session found"
-                logger.warning(details)
-                return SessionHeartbeatResultFailure(result_details=details)
-
-            details = f"Session heartbeat successful for session: {active_session_id}"
-            return SessionHeartbeatResultSuccess(result_details=details)
-        except Exception as err:
-            details = f"Failed to handle session heartbeat: {err}"
-            logger.error(details)
-            return SessionHeartbeatResultFailure(result_details=details)
-
     def handle_engine_heartbeat_request(self, request: EngineHeartbeatRequest) -> ResultPayload:
         """Handle engine heartbeat requests.
 
@@ -525,15 +388,21 @@ class GriptapeNodes(metaclass=SingletonMeta):
             workflow_info = self._get_current_workflow_info()
 
             # Get engine name
-            engine_name = GriptapeNodes.EngineIdentityManager().get_engine_name()
+            engine_name = GriptapeNodes.EngineIdentityManager().engine_name
+
+            # Get user and organization
+            user = GriptapeNodes.UserManager().user
+            user_organization = GriptapeNodes.UserManager().user_organization
 
             return EngineHeartbeatResultSuccess(
                 heartbeat_id=request.heartbeat_id,
                 engine_version=engine_version,
                 engine_name=engine_name,
-                engine_id=GriptapeNodes.EngineIdentityManager().get_active_engine_id(),
-                session_id=GriptapeNodes.SessionManager().get_active_session_id(),
+                engine_id=GriptapeNodes.EngineIdentityManager().active_engine_id,
+                session_id=GriptapeNodes.SessionManager().active_session_id,
                 timestamp=datetime.now(tz=UTC).isoformat(),
+                user=user,
+                user_organization=user_organization,
                 result_details="Engine heartbeat successful",
                 **instance_info,
                 **workflow_info,
@@ -543,40 +412,6 @@ class GriptapeNodes(metaclass=SingletonMeta):
             logger.error(details)
             return EngineHeartbeatResultFailure(heartbeat_id=request.heartbeat_id, result_details=details)
 
-    def handle_get_engine_name_request(self, request: GetEngineNameRequest) -> ResultPayload:  # noqa: ARG002
-        """Handle requests to get the current engine name."""
-        try:
-            engine_name = GriptapeNodes.EngineIdentityManager().get_engine_name()
-            return GetEngineNameResultSuccess(
-                engine_name=engine_name, result_details="Engine name retrieved successfully."
-            )
-        except Exception as err:
-            error_message = f"Failed to get engine name: {err}"
-            logger.error(error_message)
-            return GetEngineNameResultFailure(error_message=error_message, result_details=error_message)
-
-    def handle_set_engine_name_request(self, request: SetEngineNameRequest) -> ResultPayload:
-        """Handle requests to set a new engine name."""
-        try:
-            # Validate engine name (basic validation)
-            if not request.engine_name or not request.engine_name.strip():
-                error_message = "Engine name cannot be empty"
-                logger.warning(error_message)
-                return SetEngineNameResultFailure(error_message=error_message, result_details=error_message)
-
-            # Set the new engine name
-            GriptapeNodes.EngineIdentityManager().set_engine_name(request.engine_name.strip())
-            details = f"Engine name set to: {request.engine_name.strip()}"
-            return SetEngineNameResultSuccess(
-                engine_name=request.engine_name.strip(),
-                result_details=ResultDetails(message=details, level=logging.INFO),
-            )
-
-        except Exception as err:
-            error_message = f"Failed to set engine name: {err}"
-            logger.error(error_message)
-            return SetEngineNameResultFailure(error_message=error_message, result_details=error_message)
-
     def _get_instance_info(self) -> dict[str, str | None]:
         """Get instance information from environment variables.
 
